#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#define MAXBUFSIZE 100

/* Handler for SIGCHLD */

void sig_chld(int signo)
{
  pid_t pid;
  int stat; 
  while ((pid = waitpid(-1, &stat, WNOHANG)) > 0) 
    printf("child %d terminated.\n", pid);
  return ;
}



void process_request(int fd)
{
  char buf[MAXBUFSIZE];
  int num;

  while ((num = read(fd, buf, MAXBUFSIZE))) 
  {
    if (num > 0) 
      write(fd, buf, num);
  }

}


int main(int argc, char * argv[])
{
  int sockfd, rec_sock;
  socklen_t len;
  struct sigaction sa;
  char address[INET_ADDRSTRLEN];
  struct sockaddr_in addr;

  if (argc < 2) 
  {
    fprintf(stderr,"USAGE: %s port.\n", argv[0]);
    exit(-1);
  }

  /* Handle SIGCHLD */
  sa.sa_handler = sig_chld;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(SIGCHLD, &sa, NULL);

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
  {
    perror("Cannot get socket");
    exit(1);
  }

  memset(&addr, 0, sizeof(struct sockaddr_in));
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_family = AF_INET;
  addr.sin_port = htons(atoi(argv[1]));

  if (bind(sockfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) 
  {
    perror("bind");
    exit(1);
  }

  
  len = sizeof(addr);
  if (getsockname(sockfd, (struct sockaddr *)&addr, &len) < 0)
  {
    perror("can't get name");
    exit(1);
  }
  
  printf("ip = %s, port = %hd\n", inet_ntop(AF_INET, &addr.sin_addr, address, sizeof(address)), ntohs(addr.sin_port));

  if (listen(sockfd, 5) < 0) 
  {
    perror("listen");
    exit(1);
  }

  for(;;)
  {
    if( (rec_sock = accept(sockfd, NULL, NULL)) < 0)
    {
      if (errno == EINTR)
	continue;
      else 
      {
	perror("accept");
	exit(1);
      }
    }
    
    if (fork() == 0) /* Child */
    {
      close(sockfd);
      process_request(rec_sock);
      close(rec_sock);
      exit(0);
    } 
    else /* Parent */
    {
      close(rec_sock); 
    }
  }

  return 0;
}


